#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import os
import shutil
from tempfile import TemporaryDirectory
from typing import Optional

import numpy as np
from torch.utils.data import DataLoader
from torch.utils.data import Dataset
from torchvision.datasets import ImageFolder
from tqdm import tqdm

from disent.dataset.data import GroundTruthData
from disent.dataset.data._groundtruth import _DiskDataMixin
from disent.dataset.data._groundtruth import _Hdf5DataMixin
from disent.dataset.data._groundtruth__dsprites import DSpritesData
from disent.dataset.util.datafile import DataFileHashedDlGen
from disent.dataset.util.formats.hdf5 import H5Builder
from disent.util.inout.files import AtomicSaveFile
from disent.util.iters import LengthIter
from disent.util.math.random import random_choice_prng

log = logging.getLogger(__name__)


# ========================================================================= #
# load imagenet-tiny data                                                   #
# ========================================================================= #


class NumpyFolder(ImageFolder):
    def __getitem__(self, idx):
        img, cls = super().__getitem__(idx)
        return np.array(img)


def _noop(x):
    return x


def load_imagenet_tiny_data(raw_data_dir):
    # load the data
    data = NumpyFolder(os.path.join(raw_data_dir, "train"))
    data = DataLoader(
        data, batch_size=64, num_workers=min(16, os.cpu_count()), shuffle=False, drop_last=False, collate_fn=_noop
    )
    # load data - this is a bit memory inefficient doing it like this instead of with a loop into a pre-allocated array
    imgs = np.concatenate(list(tqdm(data, "loading")), axis=0)
    assert imgs.shape == (100_000, 64, 64, 3)
    return imgs


def resave_imagenet_tiny_archive(orig_zipped_file, new_save_file, overwrite=False, h5_dataset_name: str = "data"):
    """
    Convert a imagenet tiny archive to an hdf5 or numpy file depending on the file extension.
    Uncompressing the contents of the archive into a temporary directory in the same folder,
    loading the images, then converting.
    """
    _, ext = os.path.splitext(new_save_file)
    assert ext in {".npz", ".h5"}, f'unsupported save extension: {repr(ext)}, must be one of: {[".npz", ".h5"]}'
    # extract zipfile into temp dir
    with TemporaryDirectory(prefix="unzip_imagenet_tiny_", dir=os.path.dirname(orig_zipped_file)) as temp_dir:
        log.info(f"Extracting into temporary directory: {temp_dir}")
        shutil.unpack_archive(filename=orig_zipped_file, extract_dir=temp_dir)
        images = load_imagenet_tiny_data(raw_data_dir=os.path.join(temp_dir, "tiny-imagenet-200"))
    # save the data
    with AtomicSaveFile(new_save_file, overwrite=overwrite) as temp_file:
        # check the mode
        with H5Builder(temp_file, "atomic_w") as builder:
            builder.add_dataset_from_array(
                name=h5_dataset_name,
                array=images,
                chunk_shape="batch",
                compression_lvl=4,
                attrs=None,
                show_progress=True,
            )


# ========================================================================= #
# cars3d data object                                                        #
# ========================================================================= #


class ImageNetTinyDataFile(DataFileHashedDlGen):
    """
    download the cars3d dataset and convert it to a hdf5 file.
    """

    dataset_name: str = "data"

    def _generate(self, inp_file: str, out_file: str):
        resave_imagenet_tiny_archive(
            orig_zipped_file=inp_file, new_save_file=out_file, overwrite=True, h5_dataset_name=self.dataset_name
        )


class ImageNetTinyData(_Hdf5DataMixin, _DiskDataMixin, Dataset, LengthIter):
    name = "imagenet_tiny"

    datafile_imagenet_h5 = ImageNetTinyDataFile(
        uri="http://cs231n.stanford.edu/tiny-imagenet-200.zip",
        uri_hash={"fast": "4d97ff8efe3745a3bba9917d6d536559", "full": "90528d7ca1a48142e341f4ef8d21d0de"},
        file_hash={"fast": "9c23e8ec658b1ec9f3a86afafbdbae51", "full": "4c32b0b53f257ac04a3afb37e3a4204e"},
        uri_name="tiny-imagenet-200.zip",
        file_name="tiny-imagenet-200.h5",
        hash_mode="full",
    )

    datafiles = (datafile_imagenet_h5,)

    def __init__(self, data_root: Optional[str] = None, prepare: bool = False, in_memory=False, transform=None):
        super().__init__()
        self._transform = transform
        # initialize mixin
        self._mixin_disk_init(
            data_root=data_root,
            prepare=prepare,
        )
        self._mixin_hdf5_init(
            h5_path=os.path.join(self.data_dir, self.datafile_imagenet_h5.out_name),
            h5_dataset_name=self.datafile_imagenet_h5.dataset_name,
            in_memory=in_memory,
        )

    def __getitem__(self, idx: int):
        obs = self._data[idx]
        if self._transform is not None:
            obs = self._transform(obs)
        return obs


# ========================================================================= #
# dataset_dsprites                                                          #
# ========================================================================= #


class DSpritesImagenetData(GroundTruthData):
    """
    Michlo et al.
    https://github.com/nmichlo/msc-research

    DSprites that has imagenet images replacing the background or foreground.
    - This real world noise masks the ground-truth factors and
      hinders disentanglement based on the visibility [0, 100].
    - If the visibility is zero, then this is equivalent to the original
      dSprites dataset but instead with three channels.
    """

    name = "dsprites_imagenet"

    # original dsprites it only (64, 64, 1) imagenet adds the colour channel
    img_shape = (64, 64, 3)
    factor_names = DSpritesData.factor_names
    factor_sizes = DSpritesData.factor_sizes

    def __init__(
        self,
        visibility: int = 100,
        mode: str = "bg",
        data_root: Optional[str] = None,
        prepare: bool = False,
        in_memory=False,
        transform=None,
    ):
        super().__init__(transform=transform)
        # check visibility and convert to ratio
        assert isinstance(
            visibility, int
        ), f"incorrect visibility percentage type, expected int, got: {type(visibility)}"
        assert (
            0 <= visibility <= 100
        ), f"incorrect visibility percentage: {repr(visibility)}, must be in range [0, 100]. "
        self._visibility = visibility / 100
        # check mode and convert to foreground boolean
        assert mode in {"bg", "fg"}, f'incorrect mode: {repr(mode)}, must be one of: ["bg", "fg"]'
        self._foreground = mode == "fg"
        # handle the datasets
        self._dsprites = DSpritesData(data_root=data_root, prepare=prepare, in_memory=in_memory, transform=None)
        self._imagenet = ImageNetTinyData(data_root=data_root, prepare=prepare, in_memory=in_memory, transform=None)
        # deterministic randomization of the imagenet order
        self._imagenet_order = random_choice_prng(
            len(self._imagenet),
            size=len(self),
            seed=42,
        )

    def _get_observation(self, idx):
        # we need to combine the two dataset images
        # dsprites contains only {0, 255} for values
        # we can directly use these values to mask the imagenet image
        bg = self._imagenet[self._imagenet_order[idx]]
        fg = self._dsprites[idx].repeat(3, axis=-1)
        # compute background
        # set foreground
        r = self._visibility
        if self._foreground:
            # lerp content to white, and then insert into fg regions
            # r*bg + (1-r)*255
            obs = (r * bg + ((1 - r) * 255)).astype("uint8")
            obs[fg <= 127] = 0
        else:
            # lerp content to black, and then insert into bg regions
            # r*bg + (1-r)*000
            obs = (r * bg).astype("uint8")
            obs[fg > 127] = 255
        # checks
        return obs


# ========================================================================= #
# STATS                                                                     #
# ========================================================================= #


"""
dsprites_fg_100
    vis_mean: [0.02067051643494642, 0.018688392816012946, 0.01632900510079384]
    vis_std: [0.10271307751834059, 0.09390213983525653, 0.08377594259970281]
dsprites_fg_80
    vis_mean: [0.024956427531012196, 0.02336780403840578, 0.021475119672280243]
    vis_std: [0.11864125016313823, 0.11137998105649799, 0.10281424917834255]
dsprites_fg_60
    vis_mean: [0.029335176871153983, 0.028145355435322966, 0.026731731769287146]
    vis_std: [0.13663242436043319, 0.13114320478634894, 0.1246542727733097]
dsprites_fg_40
    vis_mean: [0.03369999506331255, 0.03290657349801835, 0.03196482946320608]
    vis_std: [0.155514074438101, 0.1518464537731621, 0.14750944591836743]
dsprites_fg_20
    vis_mean: [0.038064750024334834, 0.03766780505193579, 0.03719798677641122]
    vis_std: [0.17498878664096565, 0.17315570657628315, 0.1709923319496426]

dsprites_bg_100
    vis_mean: [0.5020433619489952, 0.47206398913310593, 0.42380018909780404]
    vis_std: [0.2505510666843685, 0.2500725980366869, 0.2562415603123114]
dsprites_bg_80
    vis_mean: [0.40867981393820857, 0.38468564002021527, 0.34611573047508204]
    vis_std: [0.22048328737091344, 0.22102216869942384, 0.22692977053753483]
dsprites_bg_60
    vis_mean: [0.31676960943447674, 0.29877166834408025, 0.2698556821388113]
    vis_std: [0.19745897110349, 0.1986606891520453, 0.203808842880044]
dsprites_bg_40
    vis_mean: [0.2248598986983768, 0.21285772298967615, 0.19359577132944206]
    vis_std: [0.1841631708032332, 0.18554895825833284, 0.1893568926398198]
dsprites_bg_20
    vis_mean: [0.13294969414492142, 0.12694375140936273, 0.11733572285575933]
    vis_std: [0.18311250427586276, 0.1840916474752131, 0.18607373519458442]

dsprites_fg_75
    vis_mean: [0.02606445677382044, 0.024577082627819637, 0.02280587082174753]
    vis_std: [0.12307153238282868, 0.11624914830767437, 0.1081911967745551]
dsprites_fg_50
    vis_mean: [0.031541090790578506, 0.030549541980176148, 0.029368756624861398]
    vis_std: [0.14609029304575144, 0.14150919987547408, 0.13607872227034773]
dsprites_fg_25
    vis_mean: [0.03697718115834816, 0.03648095993826591, 0.03589183623762013]
    vis_std: [0.17009317531572005, 0.16780075430655303, 0.16508779008691726]
dsprites_fg_0
    vis_mean: [0.042494423521889584, 0.042494423521889584, 0.042494423521889584]
    vis_std: [0.1951664588062605, 0.1951664588062605, 0.1951664588062605]
dsprites
    vis_mean: [0.042494423521889584]
    vis_std: [0.1951664588062605]

dsprites_bg_75
    vis_mean: [0.38577296742807327, 0.3632825822323436, 0.3271231888851156]
    vis_std: [0.21392191050784257, 0.2146731716558466, 0.2204460568339597]
dsprites_bg_50
    vis_mean: [0.271323621109491, 0.25634066038331416, 0.23223046934400662]
    vis_std: [0.18930391112143766, 0.19067969524425118, 0.19523218572886117]
dsprites_bg_25
    vis_mean: [0.15596283852200074, 0.14847876264131535, 0.13644703866118635]
    vis_std: [0.18208653250875798, 0.18323109038468802, 0.18569624396763393]
dsprites_bg_0
    vis_mean: [0.042494423521889584, 0.042494423521889584, 0.042494423521889584]
    vis_std: [0.1951664588062605, 0.1951664588062605, 0.1951664588062605]
dsprites
    vis_mean: [0.042494423521889584]
    vis_std: [0.1951664588062605]
"""


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    def compute_all_stats():
        from disent.dataset.transform import ToImgTensorF32
        from disent.dataset.util.stats import compute_data_mean_std
        from disent.util.function import wrapped_partial
        from disent.util.visualize.plot import plt_subplots_imshow

        def compute_stats(visibility: Optional[int], mode: Optional[str]):
            import psutil

            # get class
            is_imgnet = (visibility is not None) and (mode is not None)
            data_cls = (
                wrapped_partial(DSpritesImagenetData, visibility=visibility, mode=mode) if is_imgnet else DSpritesData
            )
            data_title = (
                f"{DSpritesImagenetData.name} visibility={repr(visibility)} mode={repr(mode)}"
                if is_imgnet
                else f"{DSpritesData.name}"
            )
            data_name = f"dsprites_{mode}_{visibility}" if is_imgnet else f"dsprites"
            # plot images
            data = data_cls(prepare=True)
            grid = np.array([data[i * 24733] for i in np.arange(16)]).reshape([4, 4, *data.img_shape])
            plt_subplots_imshow(grid, show=True, title=data_title)
            # compute stats
            data = data_cls(prepare=True, transform=ToImgTensorF32())
            mean, std = compute_data_mean_std(
                data, batch_size=256, num_workers=min(psutil.cpu_count(logical=False), 64), progress=True
            )
            print(f"{data_name}\n    vis_mean: {mean.tolist()}\n    vis_std: {std.tolist()}")
            # return stats
            return data_name, mean, std

        # compute common stats
        stats = []
        for mode in ["fg", "bg"]:
            for vis in [
                100,
                80,
                60,
                40,
                20,
                75,
                50,
                25,
                0,
            ]:
                stats.append(compute_stats(vis, mode))
            # 0 above should be the same as this:
            stats.append(compute_stats(None, None))

        # print once at end
        for name, mean, std in stats:
            print(f"{name}\n    vis_mean: {mean.tolist()}\n    vis_std: {std.tolist()}")

    compute_all_stats()

    # DEBUG MEMORY USAGE:

    # def debug_mem():
    #     import matplotlib.pyplot as plt
    #     from disent.dataset import DisentDataset
    #     from disent.dataset.sampling import GroundTruthDistSampler
    #     from disent.util.profiling import get_memory_usage
    #
    #     data = DSpritesImagenetData(visibility=50, in_memory=True)
    #     dataset = DisentDataset(data, sampler=GroundTruthDistSampler(num_samples=3))
    #
    #     dataloader_makers = [
    #         lambda dataset: DataLoader(dataset, batch_size=256, shuffle=True, num_workers=os.cpu_count(), pin_memory=True),
    #         lambda dataset: DataLoader(dataset, batch_size=256, shuffle=True, num_workers=os.cpu_count(), pin_memory=False),
    #     ]
    #
    #     for i, dataloader_maker in enumerate(dataloader_makers):
    #         dataloader = dataloader_maker(dataset)
    #         memory = []
    #         for j in range(1):
    #             for k, batch in enumerate(tqdm(dataloader, desc=f'{i}:{j}')):
    #                 [x.cuda() for x in batch['x_targ']]  # memory leak isnt here either...
    #                 if k % 10 == 0: memory.append(get_memory_usage())
    #                 if k % 100 == 0: tqdm.write(f'{get_memory_usage(pretty=True)}')
    #         plt.plot(range(len(memory)), memory)
    #         plt.show()
    #
    # # entry
    # debug_mem()


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import torch
from torch.distributions import Distribution

# ========================================================================= #
# Kl Losses                                                                 #
# ========================================================================= #


def kl_loss_direct_reverse(posterior: Distribution, prior: Distribution, z_sampled: torch.Tensor = None):
    # This is how the original VAE/BetaVAE papers do it.
    # - we compute the reverse kl divergence directly instead of approximating it
    # - kl(post|prior)
    # FORWARD vs. REVERSE kl (https://www.tuananhle.co.uk/notes/reverse-forward-kl.html)
    # - If we minimize the kl(post|prior) or the reverse/exclusive KL, the zero-forcing/mode-seeking behavior arises.
    # - If we minimize the kl(prior|post) or the forward/inclusive KL, the mass-covering/mean-seeking behavior arises.
    return torch.distributions.kl_divergence(posterior, prior)


def kl_loss_approx_reverse(posterior: Distribution, prior: Distribution, z_sampled: torch.Tensor = None):
    # This is how pytorch-lightning-bolts does it:
    # - kl(post|prior)
    # See issue: https://github.com/PyTorchLightning/pytorch-lightning-bolts/issues/565
    # - we approximate the reverse kl divergence instead of computing it analytically
    assert (
        z_sampled is not None
    ), 'to compute the approximate kl loss, z_sampled needs to be defined (cfg.kl_mode="approx")'
    return posterior.log_prob(z_sampled) - prior.log_prob(z_sampled)


def kl_loss_direct_forward(posterior: Distribution, prior: Distribution, z_sampled: torch.Tensor = None):
    # compute the forward kl
    # - kl(prior|post)
    return torch.distributions.kl_divergence(prior, posterior)


def kl_loss_approx_forward(posterior: Distribution, prior: Distribution, z_sampled: torch.Tensor = None):
    # compute the approximate forward kl
    # - kl(prior|post)
    assert (
        z_sampled is not None
    ), 'to compute the approximate kl loss, z_sampled needs to be defined (cfg.kl_mode="approx")'
    return prior.log_prob(z_sampled) - posterior.log_prob(z_sampled)


def kl_loss_direct_symmetric(posterior: Distribution, prior: Distribution, z_sampled: torch.Tensor = None):
    # compute the (scaled) symmetric kl
    # - 0.5 * kl(prior|post) + 0.5 * kl(prior|post)
    return 0.5 * kl_loss_direct_reverse(posterior, prior, z_sampled) + 0.5 * kl_loss_direct_forward(
        posterior, prior, z_sampled
    )


def kl_loss_approx_symmetric(posterior: Distribution, prior: Distribution, z_sampled: torch.Tensor = None):
    # compute the approximate (scaled) symmetric kl
    # - 0.5 * kl(prior|post) + 0.5 * kl(prior|post)
    return 0.5 * kl_loss_approx_reverse(posterior, prior, z_sampled) + 0.5 * kl_loss_approx_forward(
        posterior, prior, z_sampled
    )


_KL_LOSS_MODES = {
    # reverse kl -- how it should be done for VAEs
    "direct": kl_loss_direct_reverse,  # alias for reverse modes
    "approx": kl_loss_approx_reverse,  # alias for reverse modes
    "direct_reverse": kl_loss_direct_reverse,
    "approx_reverse": kl_loss_approx_reverse,
    # forward kl
    "direct_forward": kl_loss_direct_forward,
    "approx_forward": kl_loss_approx_forward,
    # symmetric kl
    "direct_symmetric": kl_loss_direct_symmetric,
    "approx_symmetric": kl_loss_approx_symmetric,
}


def kl_loss(posterior: Distribution, prior: Distribution, z_sampled: torch.Tensor = None, mode="direct"):
    return _KL_LOSS_MODES[mode](posterior, prior, z_sampled)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~


from typing import Sequence
from typing import Tuple
from typing import final

import torch
from torch.distributions import Distribution
from torch.distributions import Laplace
from torch.distributions import Normal

import disent.registry as R
from disent.frameworks.helper.util import compute_ave_loss
from disent.nn.loss.kl import kl_loss
from disent.nn.loss.reduction import loss_reduction

# ========================================================================= #
# Vae Distributions                                                         #
# TODO: this should be moved into NNs                                       #
# TODO: encoder modules should directly output distributions!               #
# ========================================================================= #


class LatentDistsHandler(object):
    def __init__(self, kl_mode: str = "direct", reduction="mean"):
        self._kl_mode = kl_mode
        self._reduction = reduction

    def encoding_to_representation(self, z_raw: Tuple[torch.Tensor, ...]) -> torch.Tensor:
        raise NotImplementedError

    def encoding_to_dists(self, z_raw: Tuple[torch.Tensor, ...]) -> Tuple[Distribution, Distribution]:
        raise NotImplementedError

    @final
    def encoding_to_dists_and_sample(
        self, z_raw: Tuple[torch.Tensor, ...]
    ) -> Tuple[Distribution, Distribution, torch.Tensor]:
        """
        Return the parameterized prior and the approximate posterior distributions,
        as well as a sample from the approximate posterior using the 'reparameterization trick'.
        """
        posterior, prior = self.encoding_to_dists(z_raw)
        # sample from posterior -- reparameterization trick!
        # ie. z ~ q(z|x)
        z_sampled = posterior.rsample()
        # return values
        return posterior, prior, z_sampled

    @final
    def compute_kl_loss(
        self, posterior: Distribution, prior: Distribution, z_sampled: torch.Tensor = None
    ) -> torch.Tensor:
        """
        Compute the kl divergence
        """
        kl = kl_loss(posterior, prior, z_sampled, mode=self._kl_mode)
        kl = loss_reduction(kl, reduction=self._reduction)
        return kl

    @final
    def compute_unreduced_kl_loss(
        self, posterior: Distribution, prior: Distribution, z_sampled: torch.Tensor = None
    ) -> torch.Tensor:
        return kl_loss(posterior, prior, z_sampled, mode=self._kl_mode)

    @final
    def compute_ave_kl_loss(
        self, ds_posterior: Sequence[Distribution], ds_prior: Sequence[Distribution], zs_sampled: Sequence[torch.Tensor]
    ) -> torch.Tensor:
        return compute_ave_loss(self.compute_kl_loss, ds_posterior, ds_prior, zs_sampled)


# ========================================================================= #
# Normal Distribution                                                       #
# ========================================================================= #


class LatentDistsHandlerNormal(LatentDistsHandler):
    """
    Latent distributions with:
    - posterior: normal distribution with diagonal covariance
    - prior: unit normal distribution

    NOTE: Expanding parameters results in something akin to an L2 regularizer.
    """

    # assert mode == 'direct', f'legacy reference implementation of KL loss only supports mode="direct", not {repr(mode)}'
    # assert reduction == 'mean_sum', f'legacy reference implementation of KL loss only supports reduction="mean_sum", not {repr(reduction)}'

    def encoding_to_representation(self, raw_z: Tuple[torch.Tensor, ...]) -> torch.Tensor:
        z_mean, z_logvar = raw_z
        return z_mean

    def encoding_to_dists(self, raw_z: Tuple[torch.Tensor, ...]) -> Tuple[Normal, Normal]:
        """
        Return the parameterized prior and the approximate posterior distributions.
        - The standard VAE parameterizes the gaussian normal with diagonal covariance.
        - logvar is used to avoid negative values for the standard deviation
        - Gaussian Encoder Model Distribution: pg. 25 in Variational Auto Encoders

        (✓) Visual inspection against reference implementations:
            https://github.com/google-research/disentanglement_lib (sample_from_latent_distribution)
            https://github.com/PyTorchLightning/pytorch-lightning-bolts/blob/master/pl_bolts/models/autoencoders/basic_vae/basic_vae_module.py
        """
        z_mean, z_logvar = raw_z
        # compute required values
        z_std = torch.exp(0.5 * z_logvar)
        # q: approximate posterior distribution
        posterior = Normal(loc=z_mean, scale=z_std)
        # p: prior distribution
        prior = Normal(torch.zeros_like(z_mean), torch.ones_like(z_std))
        # return values
        return posterior, prior


class LatentDistsHandlerLaplace(LatentDistsHandler):
    """
    Latent distributions with:
    - posterior: laplace distribution with diagonal covariance
    - prior: unit laplace distribution

    TODO: is this true?
    NOTE: Expanding parameters results in something akin to an L1 regularizer, with extra terms?
    """

    def encoding_to_representation(self, raw_z: Tuple[torch.Tensor, ...]) -> torch.Tensor:
        z_loc, z_logscale = raw_z
        return z_loc

    def encoding_to_dists(self, raw_z: Tuple[torch.Tensor, ...]) -> Tuple[Laplace, Laplace]:
        z_loc, z_logscale = raw_z
        # compute required values
        z_scale = torch.exp(z_logscale)
        # q: approximate posterior distribution
        posterior = Laplace(loc=z_loc, scale=z_scale)
        # p: prior distribution
        prior = Laplace(torch.zeros_like(z_loc), torch.ones_like(z_scale))
        # return values
        return posterior, prior


# ========================================================================= #
# Factory                                                                   #
# ========================================================================= #


def make_latent_distribution(name: str, kl_mode: str, reduction: str) -> LatentDistsHandler:
    cls = R.LATENT_HANDLERS[name]
    # make instance
    return cls(kl_mode=kl_mode, reduction=reduction)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import torch

from disent.nn.modules import DisentModule

# ========================================================================= #
# Swish                                                                     #
# ========================================================================= #


class _SwishFunction(torch.autograd.Function):
    """
    Modified from:
    https://github.com/ceshine/EfficientNet-PyTorch/blob/master/efficientnet_pytorch/utils.py
    """

    # This uses less memory than the standard implementation,
    # by re-computing the gradient on the backward pass

    @staticmethod
    def forward(ctx, x: torch.Tensor):
        y = x * torch.sigmoid(x)
        ctx.save_for_backward(x)
        return y

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor):
        x = ctx.saved_variables[0]
        sigmoid_x = torch.sigmoid(x)
        return grad_output * (sigmoid_x * (1 + x * (1 - sigmoid_x)))


def swish(tensor: torch.Tensor):
    return _SwishFunction.apply(tensor)


class Swish(DisentModule):
    def forward(self, tensor: torch.Tensor):
        return _SwishFunction.apply(tensor)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from dataclasses import dataclass
from numbers import Number
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import torch
import torchvision
from torch import Tensor
from torch.nn import functional as F
from torchvision.models import vgg19_bn

from disent.dataset.transform.functional import check_tensor
from disent.frameworks.helper.util import compute_ave_loss
from disent.frameworks.vae._unsupervised__betavae import BetaVae
from disent.nn.loss.reduction import batch_loss_reduction
from disent.nn.loss.reduction import get_mean_loss_scale

# ========================================================================= #
# Dfc Vae                                                                   #
# ========================================================================= #


class DfcVae(BetaVae):
    """
    Deep Feature Consistent Variational Autoencoder.
    https://arxiv.org/abs/1610.00291
    - Uses features generated from a pretrained model as the loss.

    Reference implementation is from: https://github.com/AntixK/PyTorch-VAE

    Difference:
        1. MSE loss changed to BCE or MSE loss
        2. Mean taken over (batch for sum of pixels) not mean over (batch & pixels)
    """

    REQUIRED_OBS = 1

    @dataclass
    class cfg(BetaVae.cfg):
        feature_layers: Optional[List[Union[str, int]]] = None
        feature_inputs_mode: str = "none"

    def __init__(self, model: "AutoEncoder", cfg: cfg = None, batch_augment=None):
        super().__init__(model=model, cfg=cfg, batch_augment=batch_augment)
        # make dfc loss
        # TODO: this should be converted to a reconstruction loss handler that wraps another handler
        self._dfc_loss = DfcLossModule(feature_layers=self.cfg.feature_layers, input_mode=self.cfg.feature_inputs_mode)

    # --------------------------------------------------------------------- #
    # Overrides                                                             #
    # --------------------------------------------------------------------- #

    def compute_ave_recon_loss(
        self, xs_partial_recon: Sequence[torch.Tensor], xs_targ: Sequence[torch.Tensor]
    ) -> Tuple[Union[torch.Tensor, Number], Dict[str, Any]]:
        # compute ave reconstruction loss
        pixel_loss = self.recon_handler.compute_ave_loss_from_partial(xs_partial_recon, xs_targ)  # (DIFFERENCE: 1)
        # compute ave deep features loss
        xs_recon = self.recon_handler.activate_all(xs_partial_recon)
        feature_loss = compute_ave_loss(
            self._dfc_loss.compute_loss, xs_recon, xs_targ, reduction=self.cfg.loss_reduction
        )
        # reconstruction error
        # TODO: not in reference implementation, but terms should be weighted
        # TODO: not in reference but feature loss is not scaled properly
        recon_loss = (pixel_loss + feature_loss) * 0.5
        # return logs
        return recon_loss, {
            "pixel_loss": pixel_loss,
            "feature_loss": feature_loss,
        }


# ========================================================================= #
# Helper Loss                                                               #
# ========================================================================= #


class DfcLossModule(torch.nn.Module):
    """
    Loss function for the Deep Feature Consistent Variational Autoencoder.
    https://arxiv.org/abs/1610.00291

    Reference implementation is from: https://github.com/AntixK/PyTorch-VAE

    Difference:
    - normalise data as torchvision.models require.

    # TODO: this should be converted to a reconstruction loss handler
    """

    def __init__(self, feature_layers: Optional[List[Union[str, int]]] = None, input_mode: str = "none"):
        """
        :param feature_layers: List of string of IDs of feature layers in pretrained model
        """
        super().__init__()
        # feature layers to use
        self.feature_layers = set(
            ["14", "24", "34", "43"] if (feature_layers is None) else [str(l) for l in feature_layers]
        )
        # feature network
        self.feature_network = vgg19_bn(pretrained=True)
        # Freeze the pretrained feature network
        for param in self.feature_network.parameters():
            param.requires_grad = False
        # Evaluation Mode
        self.feature_network.eval()
        # input node
        assert input_mode in {"none", "clamp", "assert"}
        self.input_mode = input_mode

    def compute_pairwise_loss(self, x_recon, x_targ, reduction="mean"):
        """
        THIS DOES NOT HAVE LOSS SCALING, LIKE `compute_loss`

        x_recon and x_targ data should be an unnormalized RGB batch of
        data [B x C x H x W] in the range [0, 1].
        """
        features_recon = self._extract_features(x_recon)
        features_targ = self._extract_features(x_targ)
        # compute losses
        feature_loss = 0.0
        for f_recon, f_targ in zip(features_recon, features_targ):
            loss = F.mse_loss(f_recon, f_targ, reduction="none")
            feature_loss += batch_loss_reduction(loss, reduction=reduction)
        # checks
        assert (feature_loss.ndim == 1) and (len(feature_loss) == len(x_recon))
        return feature_loss

    def compute_loss(self, x_recon, x_targ, reduction="mean"):
        """
        x_recon and x_targ data should be an unnormalized RGB batch of
        data [B x C x H x W] in the range [0, 1].
        """
        features_recon = self._extract_features(x_recon)
        features_targ = self._extract_features(x_targ)
        # compute losses
        # TODO: not in reference implementation, but consider calculating mean feature loss rather than sum
        feature_loss = 0.0
        for f_recon, f_targ in zip(features_recon, features_targ):
            feature_loss += F.mse_loss(f_recon, f_targ, reduction="mean")
        # scale the loss accordingly
        # (DIFFERENCE: 2)
        return feature_loss * get_mean_loss_scale(x_targ, reduction=reduction)

    def _extract_features(self, inputs: Tensor) -> List[Tensor]:
        """
        Extracts the features from the pretrained model
        at the layers indicated by feature_layers.
        :param inputs: (Tensor) [B x C x H x W] unnormalised in the range [0, 1].
        :return: List of the extracted features
        """
        inputs = self._process_inputs(inputs)
        # normalise: https://pytorch.org/docs/stable/torchvision/models.html
        result = torchvision.transforms.functional.normalize(
            inputs, mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
        )
        # calculate all features
        features = []
        for key, module in self.feature_network.features._modules.items():
            result = module(result)
            if key in self.feature_layers:
                features.append(result)
        return features

    def _process_inputs(self, inputs: torch.Tensor) -> torch.Tensor:
        # check the input tensor
        if self.input_mode == "assert":
            inputs = check_tensor(inputs, low=0, high=1, dtype=None)
        elif self.input_mode == "clamp":
            inputs = torch.clamp(inputs, 0, 1)
        elif self.input_mode != "none":
            raise KeyError(f"invalid input_mode={repr(self.input_mode)}")
        # repeat if missing dimensions, supports C in [1, 3]
        B, C, H, W = inputs.shape
        if C == 1:
            inputs = inputs.repeat(1, 3, 1, 1)
        assert (B, 3, H, W) == inputs.shape
        # done
        return inputs


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from typing import Callable
from typing import Optional

import torch

# ========================================================================= #
# Augment                                                                   #
# ========================================================================= #


class DisentDatasetTransform(object):
    """
    Applies transforms to batches generated from dataloaders of
    datasets from: disent.dataset.groundtruth
    """

    def __init__(
        self,
        transform: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
        transform_targ: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
    ):
        self.transform = transform
        self.transform_targ = transform_targ

    def __call__(self, batch):
        # transform inputs
        if self.transform is not None:
            if "x" not in batch:
                batch["x"] = batch["x_targ"]
            batch["x"] = _apply_transform_to_batch_dict(batch["x"], self.transform)
        # transform targets
        if self.transform_targ is not None:
            batch["x_targ"] = _apply_transform_to_batch_dict(batch["x_targ"], self.transform_targ)
        # done!
        return batch

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(transform={repr(self.transform)}, transform_targ={repr(self.transform_targ)})"
        )


def _apply_transform_to_batch_dict(batch, transform):
    if isinstance(batch, tuple):
        return tuple(transform(obs) for obs in batch)
    if isinstance(batch, list):
        return list(transform(obs) for obs in batch)
    else:
        return transform(batch)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~


from typing import Tuple

# ========================================================================= #
# Import Helper                                                             #
# ========================================================================= #


def _check_and_split_path(import_path: str) -> Tuple[str, ...]:
    segments = import_path.split(".")
    # make sure each segment is a valid python identifier
    if not all(map(str.isidentifier, segments)):
        raise ValueError(f"import path is invalid: {repr(import_path)}")
    # return the segments!
    return tuple(segments)


def import_obj(import_path: str):
    # checks
    segments = _check_and_split_path(import_path)
    # split path
    module_path, attr_name = ".".join(segments[:-1]), segments[-1]
    # import the module
    import importlib

    try:
        module = importlib.import_module(module_path)
    except Exception as e:
        raise ImportError(f"failed to import module: {repr(module_path)}") from e
    # import the attrs
    try:
        attr = getattr(module, attr_name)
    except Exception as e:
        raise ImportError(f"failed to get attribute: {repr(attr_name)} on module: {repr(module_path)}") from e
    # done
    return attr


def import_obj_partial(import_path: str, *partial_args, **partial_kwargs):
    obj = import_obj(import_path)
    # wrap the object if needed
    if partial_args or partial_kwargs:
        from disent.util.function import wrapped_partial

        obj = wrapped_partial(obj, *partial_args, **partial_kwargs)
    # done!
    return obj


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import numpy as np

from disent.dataset.data import XYObjectData
from disent.dataset.data import XYObjectShadedData
from disent.dataset.data import XYSquaresData
from disent.dataset.data import XYSquaresMinimalData

# ========================================================================= #
# TESTS                                                                     #
# ========================================================================= #


def test_xysquares_similarity():
    data_org = XYSquaresData()
    data_min = XYSquaresMinimalData()
    # check lengths
    assert len(data_org) == len(data_min)
    n = len(data_min)
    # check items
    for i in np.random.randint(0, n, size=100):
        assert np.allclose(data_org[i], data_min[i])
    # check bounds
    assert np.allclose(data_org[0], data_min[0])
    assert np.allclose(data_org[n - 1], data_min[n - 1])


def test_xyobject_similarity():
    for palette in XYObjectData.COLOR_PALETTES_3.keys():
        # create
        data0 = XYObjectData(palette=palette)
        data1 = XYObjectShadedData(palette=palette)
        assert len(data0) == len(data1)
        assert data0.factor_sizes == (*data1.factor_sizes[:-2], np.prod(data1.factor_sizes[-2:]))
        # check random
        for i in np.random.randint(len(data0), size=100):
            assert np.allclose(data0[i], data1[i])


def test_xyobject_grey_similarity():
    for palette in XYObjectData.COLOR_PALETTES_1.keys():
        # create
        data0 = XYObjectData(palette=palette, rgb=False)
        data1 = XYObjectShadedData(palette=palette, rgb=False)
        assert len(data0) == len(data1)
        assert data0.factor_sizes == (*data1.factor_sizes[:-2], np.prod(data1.factor_sizes[-2:]))
        # check random
        for i in np.random.randint(len(data0), size=100):
            assert np.allclose(data0[i], data1[i])


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import warnings
from functools import wraps
from typing import Callable
from typing import Iterator
from typing import Optional
from typing import Sequence
from typing import TypeVar
from typing import Union

import numpy as np
from torch.utils.data import Dataset
from torch.utils.data import IterableDataset
from torch.utils.data.dataloader import default_collate

from disent.dataset.data import GroundTruthData
from disent.dataset.sampling import BaseDisentSampler
from disent.dataset.sampling import SingleSampler
from disent.dataset.wrapper import WrappedDataset
from disent.util.deprecate import deprecated
from disent.util.iters import LengthIter
from disent.util.math.random import random_choice_prng

# ========================================================================= #
# Helper                                                                    #
# -- Checking if the wrapped data is an instance of GroundTruthData adds    #
#    complexity, but it means the user doesn't have to worry about handling #
#    potentially different instances of the DisentDataset class             #
# ========================================================================= #


class NotGroundTruthDataError(Exception):
    """
    This error is thrown if the wrapped dataset is not GroundTruthData
    """


T = TypeVar("T")


def groundtruth_only(func: T) -> T:
    @wraps(func)
    def wrapper(self: "DisentDataset", *args, **kwargs):
        if not self.is_ground_truth:
            raise NotGroundTruthDataError(
                f"Check `is_ground_truth` first before calling `{func.__name__}`, the dataset wrapped by {repr(self.__class__.__name__)} is not a {repr(GroundTruthData.__name__)}, instead got: {repr(self._dataset)}."
            )
        return func(self, *args, **kwargs)

    return wrapper


def wrapped_only(func):
    @wraps(func)
    def wrapper(self: "DisentDataset", *args, **kwargs):
        if not self.is_wrapped_data:
            raise NotGroundTruthDataError(
                f"Check `is_data_wrapped` first before calling `{func.__name__}`, the dataset wrapped by {repr(self.__class__.__name__)} is not a {repr(WrappedDataset.__name__)}, instead got: {repr(self._dataset)}."
            )
        return func(self, *args, **kwargs)

    return wrapper


# ========================================================================= #
# Dataset Wrapper                                                           #
# ========================================================================= #


_REF_ = object()


class DisentDataset(Dataset, LengthIter):
    def __init__(
        self,
        dataset: Union[Dataset, GroundTruthData],  # TODO: this should be renamed to data
        sampler: Optional[BaseDisentSampler] = None,
        transform: Optional[callable] = None,
        augment: Optional[callable] = None,
        return_indices: bool = False,  # doesn't really hurt performance, might as well leave enabled by default?
        return_factors: bool = False,
    ):
        super().__init__()
        # save attributes
        self._dataset = dataset
        self._sampler = SingleSampler() if (sampler is None) else sampler
        self._transform = transform
        self._augment = augment
        self._return_indices = return_indices
        self._return_factors = return_factors
        # check sampler
        assert isinstance(
            self._sampler, BaseDisentSampler
        ), f"{DisentDataset.__name__} got an invalid {BaseDisentSampler.__name__}: {type(self._sampler)}"
        # initialize sampler
        if not self._sampler.is_init:
            self._sampler.init(dataset)
        # warn if we are overriding a transform
        if self._transform is not None:
            if hasattr(dataset, "_transform") and dataset._transform:
                warnings.warn(
                    f"{DisentDataset.__name__} has transform specified as well as wrapped dataset: {dataset}, are you sure this is intended?"
                )
        # check the dataset if we are returning the factors
        if self._return_factors:
            assert isinstance(
                self._dataset, GroundTruthData
            ), f"If `return_factors` is `True`, then the dataset must be an instance of: {GroundTruthData.__name__}, got: {type(dataset)}"

    def shallow_copy(
        self,
        dataset: Union[Dataset, GroundTruthData] = _REF_,  # TODO: this should be renamed to data
        sampler: Optional[BaseDisentSampler] = _REF_,
        transform: Optional[callable] = _REF_,
        augment: Optional[callable] = _REF_,
        return_indices: bool = _REF_,
        return_factors: bool = _REF_,
    ) -> "DisentDataset":
        # instantiate shallow dataset copy, overwriting elements if specified
        return DisentDataset(
            dataset=self._dataset if (dataset is _REF_) else dataset,
            sampler=self._sampler.uninit_copy() if (sampler is _REF_) else sampler,
            transform=self._transform if (transform is _REF_) else transform,
            augment=self._augment if (augment is _REF_) else augment,
            return_indices=self._return_indices if (return_indices is _REF_) else return_indices,
            return_factors=self._return_factors if (return_factors is _REF_) else return_factors,
        )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Properties                                                            #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @property
    def data(self) -> Dataset:
        return self._dataset

    @property
    def sampler(self) -> BaseDisentSampler:
        return self._sampler

    @property
    def transform(self) -> Optional[Callable[[object], object]]:
        return self._transform

    @property
    def augment(self) -> Optional[Callable[[object], object]]:
        return self._augment

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Ground Truth Only                                                     #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @property
    def is_ground_truth(self) -> bool:
        return isinstance(self._dataset, GroundTruthData)

    @property
    @deprecated("ground_truth_data property replaced with `gt_data`")
    @groundtruth_only
    def ground_truth_data(self) -> GroundTruthData:
        return self._dataset

    @property
    @groundtruth_only
    def gt_data(self) -> GroundTruthData:
        # TODO: deprecate this or the long version
        return self._dataset

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Wrapped Dataset                                                       #
    # -- TODO: this is a bit hacky                                          #
    # -- Allows us to compute disentanglement metrics over datasets         #
    #    derived from ground truth data                                     #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @property
    def is_wrapped_data(self):
        return isinstance(self._dataset, WrappedDataset)

    @property
    def is_wrapped_gt_data(self):
        return isinstance(self._dataset, WrappedDataset) and isinstance(self._dataset.data, GroundTruthData)

    @property
    @wrapped_only
    def wrapped_data(self):
        self._dataset: WrappedDataset
        return self._dataset.data

    @property
    @wrapped_only
    def wrapped_gt_data(self):
        self._dataset: WrappedDataset
        return self._dataset.gt_data

    @wrapped_only
    def unwrapped_shallow_copy(
        self,
        sampler: Optional[BaseDisentSampler] = _REF_,
        transform: Optional[callable] = _REF_,
        augment: Optional[callable] = _REF_,
        return_indices: bool = _REF_,
        return_factors: bool = _REF_,
    ) -> "DisentDataset":
        # like shallow_copy, but unwrap the dataset instead!
        return self.shallow_copy(
            dataset=self.wrapped_data,
            sampler=sampler,
            transform=transform,
            augment=augment,
            return_indices=return_indices,
            return_factors=return_factors,
        )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Dataset                                                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def __len__(self):
        return len(self._dataset)

    def __getitem__(self, idx):
        if self._sampler is not None:
            idxs = self._sampler(idx)
        else:
            idxs = (idx,)
        # get the observations
        return self._dataset_get_observation(*idxs)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Single Datapoints                                                     #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def _datapoint_raw_to_target(self, dat):
        x_targ = dat
        if self._transform is not None:
            x_targ = self._transform(x_targ)
        return x_targ

    def _datapoint_target_to_input(self, x_targ):
        x = x_targ
        if self._augment is not None:
            x = self._augment(x)
            # some augmentations may convert a (C, H, W) to (1, C, H, W), undo this change
            # TODO: this should not be here! this should be handled by the user instead!
            x = _batch_to_observation(batch=x, obs_shape=x_targ.shape)
        return x

    def dataset_get(self, idx, mode: str):
        """
        Gets the specified datapoint, using the specified mode.
        - raw: direct untransformed/unaugmented observations
        - target: transformed observations
        - input: transformed then augmented observations
        - pair: (input, target) tuple of observations

        Pipeline:
            1. raw    = dataset[idx]
            2. target = transform(raw)
            3. input  = augment(target) = augment(transform(raw))

        :param idx: The index of the datapoint in the dataset
        :param mode: {'raw', 'target', 'input', 'pair'}
        :return: observation depending on mode
        """
        try:
            idx = int(idx)
        except:
            raise TypeError(f"Indices must be integer-like ({type(idx)}): {idx}")
        # we do not support indexing by lists
        x_raw = self._dataset[idx]
        # return correct data
        if mode == "pair":
            x_targ = self._datapoint_raw_to_target(x_raw)  # applies self.transform
            x = self._datapoint_target_to_input(x_targ)  # applies self.augment
            return x, x_targ
        elif mode == "input":
            x_targ = self._datapoint_raw_to_target(x_raw)  # applies self.transform
            x = self._datapoint_target_to_input(x_targ)  # applies self.augment
            return x
        elif mode == "target":
            x_targ = self._datapoint_raw_to_target(x_raw)  # applies self.transform
            return x_targ
        elif mode == "raw":
            return x_raw
        else:
            raise ValueError(f"Invalid {mode=}")

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Multiple Datapoints                                                   #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def _dataset_get_observation(self, *idxs):
        xs, xs_targ = zip(*(self.dataset_get(idx, mode="pair") for idx in idxs))
        # handle cases
        obs = {"x_targ": xs_targ}
        # 5-10% faster
        if self._augment is not None:
            obs["x"] = xs
        # add indices
        if self._return_indices:
            obs["idx"] = idxs
        # add factors
        if self._return_factors:
            # >>> this is about 10% faster than below, because we do not need to do conversions!
            obs["factors"] = tuple(np.array(np.unravel_index(idxs, self._dataset.factor_sizes)).T)
            # >>> builtin but slower method, does some magic for more than 2 dims, could replace with faster try_njit method, but then we need numba!
            # obs['factors1'] = tuple(self.gt_data.idx_to_pos(idxs))
        # done!
        return obs

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Batches                                                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    # TODO: default_collate should be replaced with a function
    #      that can handle tensors and nd.arrays, and return accordingly

    def dataset_batch_from_indices(self, indices: Sequence[int], mode: str, collate: bool = True):
        """Get a batch of observations X from a batch of factors Y."""
        batch = [self.dataset_get(idx, mode=mode) for idx in indices]
        return default_collate(batch) if collate else batch

    def dataset_sample_batch(
        self,
        num_samples: int,
        mode: str,
        replace: bool = False,
        return_indices: bool = False,
        collate: bool = True,
        seed: Optional[int] = None,
    ):
        """Sample a batch of observations X."""
        # built in np.random.choice cannot handle large values: https://github.com/numpy/numpy/issues/5299#issuecomment-497915672
        indices = random_choice_prng(len(self._dataset), size=num_samples, replace=replace, seed=seed)
        # return batch
        batch = self.dataset_batch_from_indices(indices, mode=mode, collate=collate)
        # return values
        if return_indices:
            return batch, (default_collate(indices) if collate else indices)
        else:
            return batch

    def dataset_sample_elems(
        self, num_samples: int, mode: str, return_indices: bool = False, seed: Optional[int] = None
    ):
        """Sample uncollated elements with replacement, like `dataset_sample_batch`"""
        return self.dataset_sample_batch(
            num_samples=num_samples, mode=mode, replace=True, return_indices=return_indices, collate=False, seed=seed
        )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Batches -- Ground Truth Only                                          #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    # TODO: batches should be obtained from indices
    #       - the wrapped gt datasets should handle generating these indices, eg. factor traversals etc.

    @groundtruth_only
    def dataset_batch_from_factors(self, factors: np.ndarray, mode: str, collate: bool = True):
        """Get a batch of observations X from a batch of factors Y."""
        indices = self.gt_data.pos_to_idx(factors)
        return self.dataset_batch_from_indices(indices, mode=mode, collate=collate)

    @groundtruth_only
    def dataset_sample_batch_with_factors(self, num_samples: int, mode: str, collate: bool = True):
        """Sample a batch of observations X and factors Y."""
        factors = self.gt_data.sample_factors(num_samples)
        batch = self.dataset_batch_from_factors(factors, mode=mode, collate=collate)
        return batch, (default_collate(factors) if collate else factors)


class DisentIterDataset(IterableDataset, DisentDataset):
    # make sure we cannot obtain the length directly
    __len__ = None

    def __iter__(self):
        # this takes priority over __getitem__, otherwise __getitem__ would need to
        # raise an IndexError if out of bounds to signal the end of iteration
        while True:
            # yield the entire dataset
            # - repeating when it is done!
            yield from (self[i] for i in range(len(self._dataset)))


# ========================================================================= #
# util                                                                      #
# ========================================================================= #


def _batch_to_observation(batch, obs_shape):
    """
    Convert a batch of size 1, to a single observation.
    """
    if batch.shape != obs_shape:
        assert batch.shape == (
            1,
            *obs_shape,
        ), f"batch.shape={repr(batch.shape)} does not correspond to obs_shape={repr(obs_shape)} with batch dimension added"
        return batch.reshape(obs_shape)
    return batch


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

"""
Flatness Metric
- Michlo et al.
  https://github.com/nmichlo/msc-research
"""

import logging
import math
from typing import Iterable
from typing import Tuple
from typing import Union

import torch
from torch.utils.data.dataloader import default_collate

from disent.dataset import DisentDataset
from disent.metrics.utils import make_metric
from disent.util.deprecate import deprecated
from disent.util.iters import iter_chunks

log = logging.getLogger(__name__)


# ========================================================================= #
# flatness                                                                  #
# ========================================================================= #


@make_metric("flatness", fast_kwargs=dict(repeats=128))
def metric_flatness(
    dataset: DisentDataset,
    representation_function: callable,
    repeats: int = 1024,
    batch_size: int = 64,
):
    """
    Michlo et al.
    https://github.com/nmichlo/msc-research

    Computes the flatness metric:
        approximately equal to: total_dim_width / (ave_point_dist_along_dim * num_points_along_dim)

    Complexity of this metric is:
        O(num_factors * ave_factor_size * repeats)
        eg. 9 factors * 64 indices on ave * 128 repeats = 73728 observations loaded from the dataset

    repeats:
      - can go all the way down to about 64 and still get decent results.
      - 64 is accurate to about +- 0.01
      - 128 is accurate to about +- 0.003
      - 1024 is accurate to about +- 0.001

    Args:
      dataset: DisentDataset to be sampled from.
      representation_function: Function that takes observations as input and outputs a dim_representation sized representation for each observation.
      repeats: how many times to repeat a traversal along each factors, these are then averaged together.
      batch_size: Batch size to process at any time while generating representations, should not effect metric results.
      p: how to calculate distances in the latent space, see torch.norm
    Returns:
      Dictionary with average disentanglement score, completeness and
        informativeness (train and test).
    """
    p_fs_measures = aggregate_measure_distances_along_all_factors(
        dataset, representation_function, repeats=repeats, batch_size=batch_size, ps=(1, 2)
    )
    # get info
    factor_sizes = dataset.gt_data.factor_sizes
    # aggregate data
    results = {
        # main metric -- also measures axis-alignment
        "flatness.ave_flatness": compute_flatness(
            widths=p_fs_measures[2]["fs_ave_widths"],
            lengths=p_fs_measures[1]["fs_ave_lengths"],
            factor_sizes=factor_sizes,
        ),
        # optional metrics
        "flatness.ave_flatness_l1": compute_flatness(
            widths=p_fs_measures[1]["fs_ave_widths"],
            lengths=p_fs_measures[1]["fs_ave_lengths"],
            factor_sizes=factor_sizes,
        ),
        "flatness.ave_flatness_l2": compute_flatness(
            widths=p_fs_measures[2]["fs_ave_widths"],
            lengths=p_fs_measures[2]["fs_ave_lengths"],
            factor_sizes=factor_sizes,
        ),
        # distances
        "flatness.ave_width_l1": torch.mean(
            filter_inactive_factors(p_fs_measures[1]["fs_ave_widths"], factor_sizes=factor_sizes)
        ),
        "flatness.ave_width_l2": torch.mean(
            filter_inactive_factors(p_fs_measures[2]["fs_ave_widths"], factor_sizes=factor_sizes)
        ),
        "flatness.ave_length_l1": torch.mean(
            filter_inactive_factors(p_fs_measures[1]["fs_ave_lengths"], factor_sizes=factor_sizes)
        ),
        "flatness.ave_length_l2": torch.mean(
            filter_inactive_factors(p_fs_measures[2]["fs_ave_lengths"], factor_sizes=factor_sizes)
        ),
    }
    # convert values from torch
    return {k: float(v) for k, v in results.items()}


def compute_flatness(widths, lengths, factor_sizes):
    widths = filter_inactive_factors(widths, factor_sizes)
    lengths = filter_inactive_factors(lengths, factor_sizes)
    # checks
    assert torch.all(widths >= 0)
    assert torch.all(lengths >= 0)
    assert torch.all(torch.eq(widths == 0, lengths == 0))
    # update scores
    widths[lengths == 0] = 0
    lengths[lengths == 0] = 1
    # compute flatness
    return (widths / lengths).mean()


def filter_inactive_factors(tensor, factor_sizes):
    factor_sizes = torch.tensor(factor_sizes, device=tensor.device)
    assert torch.all(factor_sizes >= 1)
    # remove
    active_factors = torch.nonzero(factor_sizes - 1, as_tuple=True)
    return tensor[active_factors]


def aggregate_measure_distances_along_all_factors(
    dataset: DisentDataset,
    representation_function,
    repeats: int,
    batch_size: int,
    ps: Iterable[Union[str, int]] = (1, 2),
) -> dict:
    # COMPUTE AGGREGATES FOR EACH FACTOR
    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
    fs_p_measures = [
        aggregate_measure_distances_along_factor(
            dataset, representation_function, f_idx=f_idx, repeats=repeats, batch_size=batch_size, ps=ps
        )
        for f_idx in range(dataset.gt_data.num_factors)
    ]

    # FINALIZE FOR EACH FACTOR
    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
    p_fs_measures = {}
    for p, fs_measures in default_collate(fs_p_measures).items():
        fs_ave_widths = fs_measures["ave_width"]
        # get number of spaces deltas (number of points minus 1)
        # compute length: estimated version of factors_ave_width = factors_num_deltas * factors_ave_delta
        _fs_num_deltas = torch.as_tensor(dataset.gt_data.factor_sizes, device=fs_ave_widths.device) - 1
        _fs_ave_deltas = fs_measures["ave_delta"]
        fs_ave_lengths = _fs_num_deltas * _fs_ave_deltas
        # update
        p_fs_measures[p] = {"fs_ave_widths": fs_ave_widths, "fs_ave_lengths": fs_ave_lengths}
    return p_fs_measures


def aggregate_measure_distances_along_factor(
    dataset: DisentDataset,
    representation_function,
    f_idx: int,
    repeats: int,
    batch_size: int,
    ps: Iterable[Union[str, int]] = (1, 2),
    cycle_fail: bool = False,
) -> dict:
    f_size = dataset.gt_data.factor_sizes[f_idx]

    if f_size == 1:
        if cycle_fail:
            raise ValueError(
                f"dataset factor size is too small for flatness metric with cycle_normalize enabled! size={f_size} < 2"
            )
        zero = torch.as_tensor(0.0, device=get_device(dataset, representation_function))
        return {p: {"ave_width": zero.clone(), "ave_delta": zero.clone()} for p in ps}

    # FEED FORWARD, COMPUTE ALL DELTAS & WIDTHS - For each distance measure
    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
    p_measures: list = [{} for _ in range(repeats)]
    for measures in p_measures:
        # generate repeated factors, varying one factor over the entire range
        zs_traversal = encode_all_along_factor(dataset, representation_function, f_idx=f_idx, batch_size=batch_size)
        # for each distance measure compute everything
        # - width: calculate the distance between the furthest two points
        # - deltas: calculating the distances of their representations to the next values.
        # - cycle_normalize: we cant get the ave next dist directly because of cycles, so we remove the largest dist
        for p in ps:
            deltas_next = torch.norm(
                torch.roll(zs_traversal, -1, dims=0) - zs_traversal, dim=-1, p=p
            )  # next | shape: (factor_size, z_size)
            deltas_prev = torch.norm(
                torch.roll(zs_traversal, 1, dims=0) - zs_traversal, dim=-1, p=p
            )  # prev | shape: (factor_size, z_size)
            # values needed for flatness
            width = knn(x=zs_traversal, y=zs_traversal, k=1, largest=True, p=p).values.max()  # shape: (,)
            min_deltas = torch.topk(
                deltas_next, k=f_size - 1, dim=-1, largest=False, sorted=False
            )  # shape: (factor_size-1, z_size)
            # TODO: other measures can be added:
            #       1. multivariate skewness
            #       2. normality measure
            #       3. independence
            #       4. menger curvature (Cayley-Menger Determinant?)
            # save variables
            measures[p] = {"widths": width, "deltas": min_deltas.values}

    # AGGREGATE DATA - For each distance measure
    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
    return {
        p: {
            "ave_width": measures["widths"].mean(dim=0),  # shape: (repeats,) -> ()
            "ave_delta": measures["deltas"].mean(dim=[0, 1]),  # shape: (repeats, factor_size - 1) -> ()
        }
        for p, measures in default_collate(p_measures).items()
    }


# ========================================================================= #
# ENCODE                                                                    #
# ========================================================================= #


def encode_all_along_factor(
    dataset: DisentDataset, representation_function, f_idx: int, batch_size: int, return_batch: bool = False
):
    # generate repeated factors, varying one factor over a range (f_size, f_dims)
    factors = dataset.gt_data.sample_random_factor_traversal(f_idx=f_idx)
    # get the representations of all the factors (f_size, z_size)
    # * if return_batch is False: return sequential_zs
    # * if return_batch is True: return (sequential_zs, sequential_batch)
    return encode_all_factors(
        dataset, representation_function, factors=factors, batch_size=batch_size, return_batch=return_batch
    )


def encode_all_factors(
    dataset: DisentDataset, representation_function, factors, batch_size: int, return_batch: bool = False
) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
    zs = []
    xs = []
    with torch.no_grad():
        for batch_factors in iter_chunks(factors, chunk_size=batch_size):
            batch = dataset.dataset_batch_from_factors(batch_factors, mode="input")
            z = representation_function(batch)
            zs.append(z)
            if return_batch:
                xs.append(batch)
    # handle case
    if return_batch:
        # might be on different devices
        return torch.cat(zs, dim=0), torch.cat(xs, dim=0)
    else:
        return torch.cat(zs, dim=0)


def get_device(dataset: DisentDataset, representation_function):
    # this is a hack...
    return representation_function(dataset.dataset_sample_batch(1, mode="input")).device


# ========================================================================= #
# DISTANCES                                                                 #
# ========================================================================= #


def knn(x, y, k: int = None, largest=False, p="fro"):
    assert 0 < k <= y.shape[0]
    # check input vectors, must be array of vectors
    assert 2 == x.ndim == y.ndim
    assert x.shape[1:] == y.shape[1:]
    # compute distances between each and every pair
    dist_mat = x[:, None, ...] - y[None, :, ...]
    dist_mat = torch.norm(dist_mat, dim=-1, p=p)
    # return closest distances
    return torch.topk(dist_mat, k=k, dim=-1, largest=largest, sorted=True)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #


# if __name__ == '__main__':
#     import lightning as L
#     from torch.optim import Adam
#     from torch.utils.data import DataLoader
#     from disent.data.groundtruth import XYObjectData, XYSquaresData
#     from disent.dataset.groundtruth import GroundTruthDataset, GroundTruthDatasetPairs
#     from disent.frameworks.vae import BetaVae
#     from disent.frameworks.vae import AdaVae
#     from disent.model.ae import EncoderConv64, DecoderConv64, AutoEncoder
#     from disent.transform import ToImgTensorF32
#     from disent.util import colors
#     from disent.util import Timer
#
#     def get_str(r):
#         return ', '.join(f'{k}={v:6.4f}' for k, v in r.items())
#
#     def print_r(name, steps, result, clr=colors.lYLW, t: Timer = None):
#         print(f'{clr}{name:<13} ({steps:>04}){f" {colors.GRY}[{t.pretty}]{clr}" if t else ""}: {get_str(result)}{colors.RST}')
#
#     def calculate(name, steps, dataset, get_repr):
#         global aggregate_measure_distances_along_factor
#         with Timer() as t:
#             r = metric_flatness(dataset, get_repr, repeats=64, batch_size=64)
#         results.append((name, steps, r))
#         print_r(name, steps, r, colors.lRED, t=t)
#         print(colors.GRY, '='*100, colors.RST, sep='')
#         return r
#
#     class XYOverlapData(XYSquaresData):
#         def __init__(self, square_size=8, image_size=64, grid_spacing=None, num_squares=3, rgb=True):
#             if grid_spacing is None:
#                 grid_spacing = (square_size+1) // 2
#             super().__init__(square_size=square_size, image_size=image_size, grid_spacing=grid_spacing, num_squares=num_squares, rgb=rgb)
#
#     # datasets = [XYObjectData(rgb=False, palette='white'), XYSquaresData(), XYOverlapData(), XYObjectData()]
#     datasets = [XYObjectData()]
#
#     results = []
#     for data in datasets:
#         dataset = GroundTruthDatasetPairs(data, transform=ToImgTensorF32())
#         dataloader = DataLoader(dataset=dataset, batch_size=32, shuffle=True, pin_memory=True)
#         module = AdaVae(
#             model=AutoEncoder(
#                 encoder=EncoderConv64(x_shape=data.x_shape, z_size=6, z_multiplier=2),
#                 decoder=DecoderConv64(x_shape=data.x_shape, z_size=6),
#             ),
#             cfg=AdaVae.cfg(beta=0.001, loss_reduction='mean', optimizer=torch.optim.Adam, optimizer_kwargs=dict(lr=5e-4))
#         )
#         # we cannot guarantee which device the representation is on
#         get_repr = lambda x: module.encode(x.to(module.device))
#         # PHASE 1, UNTRAINED
#         L.Trainer(logger=False, enable_checkpointing=False, fast_dev_run=True, gpus=1, weights_summary=None).fit(module, dataloader)
#         module = module.to('cuda')
#         calculate(data.__class__.__name__, 0, dataset, get_repr)
#         # PHASE 2, LITTLE TRAINING
#         L.Trainer(logger=False, enable_checkpointing=False, max_steps=256, gpus=1, weights_summary=None).fit(module, dataloader)
#         calculate(data.__class__.__name__, 256, dataset, get_repr)
#         # PHASE 3, MORE TRAINING
#         L.Trainer(logger=False, enable_checkpointing=False, max_steps=2048, gpus=1, weights_summary=None).fit(module, dataloader)
#         calculate(data.__class__.__name__, 256+2048, dataset, get_repr)
#         results.append(None)
#
#     for result in results:
#         if result is None:
#             print()
#             continue
#         (name, steps, result) = result
#         print_r(name, steps, result, colors.lYLW)



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from typing import Optional
from typing import Sequence

import torch

import disent.dataset.transform.functional as F_d

# ========================================================================= #
# Transforms                                                                #
# ========================================================================= #
from disent.util.deprecate import deprecated


class Noop(object):
    """
    Transform that does absolutely nothing!

    See: disent.transform.functional.noop
    """

    def __call__(self, obs):
        return obs

    def __repr__(self):
        return f"{self.__class__.__name__}()"


class CheckTensor(object):
    """
    Check that the data is a tensor, the right dtype, and in the required range.

    See: disent.transform.functional.check_tensor
    """

    def __init__(
        self,
        low: Optional[float] = 0.0,
        high: Optional[float] = 1.0,
        dtype: Optional[torch.dtype] = torch.float32,
    ):
        self._low = low
        self._high = high
        self._dtype = dtype

    def __call__(self, obs):
        return F_d.check_tensor(obs, low=self._low, high=self._high, dtype=self._dtype)

    def __repr__(self):
        kwargs = dict(low=self._low, high=self._high, dtype=self._dtype)
        kwargs = ", ".join(f"{k}={repr(v)}" for k, v in kwargs.items() if (v is not None))
        return f"{self.__class__.__name__}({kwargs})"


class ToImgTensorF32(object):
    """
    Basic transform that should be applied to most datasets, making sure
    the image tensor is float32 and a specified size.

    Steps:
        1. resize image if size is specified
        2. if we have integer inputs, divide by 255
        3. add missing channel to greyscale image
        4. move channels to first dim (H, W, C) -> (C, H, W)
        5. normalize using mean and std, values might thus be outside of the range [0, 1]

    See: disent.transform.functional.to_img_tensor_f32
    """

    def __init__(
        self,
        size: Optional[F_d.SizeType] = None,
        mean: Optional[Sequence[float]] = None,
        std: Optional[Sequence[float]] = None,
    ):
        self._size = size
        self._mean = tuple(mean) if (mean is not None) else None
        self._std = tuple(std) if (std is not None) else None

    def __call__(self, obs) -> torch.Tensor:
        return F_d.to_img_tensor_f32(obs, size=self._size, mean=self._mean, std=self._std)

    def __repr__(self):
        kwargs = dict(size=self._size, mean=self._mean, std=self._std)
        kwargs = ", ".join(f"{k}={repr(v)}" for k, v in kwargs.items() if (v is not None))
        return f"{self.__class__.__name__}({kwargs})"


class ToImgTensorU8(object):
    """
    Basic transform that makes sure the image tensor is uint8 and a specified size.

    Steps:
    1. resize image if size is specified
    2. add missing channel to greyscale image
    3. move channels to first dim (H, W, C) -> (C, H, W)

    See: disent.transform.functional.to_img_tensor_u8
    """

    def __init__(
        self,
        size: Optional[F_d.SizeType] = None,
    ):
        self._size = size

    def __call__(self, obs) -> torch.Tensor:
        return F_d.to_img_tensor_u8(obs, size=self._size)

    def __repr__(self):
        kwargs = f"size={repr(self._size)}" if (self._size is not None) else ""
        return f"{self.__class__.__name__}({kwargs})"


# ========================================================================= #
# Deprecated                                                                #
# ========================================================================= #


@deprecated("ToStandardisedTensor renamed to ToImgTensorF32")
class ToStandardisedTensor(ToImgTensorF32):
    pass


@deprecated("ToUint8Tensor renamed to ToImgTensorU8")
class ToUint8Tensor(ToImgTensorU8):
    pass


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from typing import List
from typing import Tuple

import numpy as np
from torch.utils.data import Dataset

from disent.util.iters import LengthIter

# ========================================================================= #
# custom episodes -- base                                                   #
# ========================================================================= #


class BaseEpisodesData(Dataset, LengthIter):
    def __init__(self, transform=None):
        self._transform = transform
        # load episodes
        self._episodes = self._load_episode_observations()
        assert len(self._episodes) > 0, "There must be at least one episode!"
        # total length
        self._lengths = np.array([len(episode) for episode in self._episodes])
        self._length = np.sum(self._lengths)
        self._weights = self._lengths / self._length

    def __len__(self):
        return self._length

    def __getitem__(self, idx):
        # this can be slow!
        # linear search is conducted!
        episode, idx, _ = self.get_episode_and_idx(idx)
        obs = episode[idx]
        if self._transform is not None:
            obs = self._transform(obs)
        return obs

    def get_episode_and_idx(self, idx) -> Tuple[np.ndarray, int, int]:
        assert idx >= 0, "Negative indices are not supported."
        # linear search for episode & shift idx accordingly
        # TODO: This could be better...
        # TODO: add caching?
        episode, offset = None, 0
        for episode in self._episodes:
            length = len(episode)
            if idx < length:
                break
            else:
                offset += length
                idx -= length
        # return found
        return episode, idx, offset

    def _load_episode_observations(self) -> List[np.ndarray]:
        raise NotImplementedError


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# CHANGES:
# - converted from tensorflow to pytorch
# - removed gin config
# - uses disent objects and classes
# - renamed functions

"""
Utility functions that are useful for the different metrics.
"""

from numbers import Number
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generic
from typing import Optional
from typing import Protocol
from typing import TypeVar
from typing import Union

import numpy as np
import sklearn
from tqdm import tqdm

from disent.dataset import DisentDataset
from disent.util import to_numpy
from disent.util.function import wrapped_partial

# ========================================================================= #
# Metric Wrapper                                                            #
# ========================================================================= #


T = TypeVar("T")


class Metric(Generic[T]):
    def __init__(
        self,
        name: str,
        metric_fn: T,  # Callable[[...], Dict[str, Number]]
        default_kwargs: Optional[Dict[str, Any]] = None,
        fast_kwargs: Optional[Dict[str, Any]] = None,
    ):
        self._name = name
        self._orig_fn = metric_fn
        self._metric_fn_default = wrapped_partial(self._orig_fn, **(default_kwargs if default_kwargs else {}))
        self._metric_fn_fast = wrapped_partial(self._orig_fn, **(fast_kwargs if fast_kwargs else {}))

    # How do we get a type hint for `__call__` so that its signature matches `T`?
    def __call__(self, *args, **kwargs) -> Dict[str, Number]:
        return self._metric_fn_default(*args, **kwargs)

    @property
    def compute(self) -> T:
        return self._metric_fn_default

    @property
    def compute_fast(self) -> T:
        return self._metric_fn_fast

    @property
    def unwrap(self) -> T:
        return self._orig_fn

    @property
    def name(self) -> str:
        return self._name

    def __str__(self):
        return f"metric-{self.name}"


def make_metric(
    name: str,
    default_kwargs: Optional[Dict[str, Any]] = None,
    fast_kwargs: Optional[Dict[str, Any]] = None,
) -> Callable[[T], Union[Metric[T], T]]:
    """
    Metrics should be decorated using this function to set defaults!
    Two versions of the metric should exist.
        1. Recommended settings
           - This should give reliable results, but may be very slow, multiple minutes to half an
             hour or more for some metrics depending on the underlying model, data and ground-truth factors.
        2. Faster settings
           - This should give a decent results, but should be decently fast, a few seconds/minutes at most.
             This is not used for testing
    """

    # `Union[Metric[T], T]` is hack to get type hint on `__call__`
    def _wrap_fn_as_metric(metric_fn: T) -> Union[Metric[T], T]:
        return Metric(name=name, metric_fn=metric_fn, default_kwargs=default_kwargs, fast_kwargs=fast_kwargs)

    return _wrap_fn_as_metric


# ========================================================================= #
# utils                                                                     #
# ========================================================================= #


def generate_batch_factor_code(
    dataset: DisentDataset,
    representation_function,
    num_points: int,
    batch_size: int,
    show_progress: bool = False,
):
    """Sample a single training sample based on a mini-batch of ground-truth data.
    Args:
      dataset: DisentDataset to be sampled from.
      representation_function: Function that takes observation as input and outputs a representation.
      num_points: Number of points to sample.
      batch_size: Batchsize to sample points.
      show_progress: if a progress bar should be shown
    Returns:
      representations: Codes (num_codes, num_points)-np array.
      factors: Factors generating the codes (num_factors, num_points)-np array.
    """
    # TODO: this can be cleaned up and simplified
    #       maybe use chunked()
    representations = None
    factors = None
    i = 0
    with tqdm(total=num_points, disable=not show_progress) as bar:
        while i < num_points:
            num_points_iter = min(num_points - i, batch_size)
            current_observations, current_factors = dataset.dataset_sample_batch_with_factors(
                num_points_iter, mode="input"
            )
            if i == 0:
                factors = current_factors
                representations = to_numpy(representation_function(current_observations))
            else:
                factors = np.vstack((factors, current_factors))
                representations = np.vstack((representations, to_numpy(representation_function(current_observations))))
            i += num_points_iter
            bar.update(num_points_iter)
    return np.transpose(representations), np.transpose(factors)


def split_train_test(observations, train_percentage):
    """
    Splits observations into a train and test set.
    Args:
      observations: Observations to split in train and test. They can be the
        representation or the observed factors of variation. The shape is
        (num_dimensions, num_points) and the split is over the points.
      train_percentage: Fraction of observations to be used for training.
    Returns:
      observations_train: Observations to be used for training.
      observations_test: Observations to be used for testing.
    """
    num_labelled_samples = observations.shape[1]
    num_labelled_samples_train = int(np.ceil(num_labelled_samples * train_percentage))
    num_labelled_samples_test = num_labelled_samples - num_labelled_samples_train
    observations_train = observations[:, :num_labelled_samples_train]
    observations_test = observations[:, num_labelled_samples_train:]
    assert observations_test.shape[1] == num_labelled_samples_test, "Wrong size of the test set."
    return observations_train, observations_test


def obtain_representation(observations, representation_function, batch_size):
    """ "Obtain representations from observations.
    Args:
      observations: Observations for which we compute the representation.
      representation_function: Function that takes observation as input and
        outputs a representation.
      batch_size: Batch size to compute the representation.
    Returns:
      representations: Codes (num_codes, num_points)-Numpy array.
    """
    representations = None
    # TODO: use chunked
    num_points = observations.shape[0]
    i = 0
    while i < num_points:
        num_points_iter = min(num_points - i, batch_size)
        current_observations = observations[i : i + num_points_iter]
        if i == 0:
            representations = to_numpy(representation_function(current_observations))
        else:
            representations = np.vstack((representations, to_numpy(representation_function(current_observations))))
        i += num_points_iter
    return np.transpose(representations)


def histogram_discretize(target, num_bins=20):
    """
    Discretization based on histograms.
    """
    discretized = np.zeros_like(target)
    for i in range(target.shape[0]):
        discretized[i, :] = np.digitize(target[i, :], np.histogram(target[i, :], num_bins)[1][:-1])
    return discretized


def discrete_mutual_info(mus, ys):
    """
    Compute discrete mutual information.
    """
    num_codes = mus.shape[0]
    num_factors = ys.shape[0]
    m = np.zeros([num_codes, num_factors])
    for i in range(num_codes):
        for j in range(num_factors):
            m[i, j] = sklearn.metrics.mutual_info_score(ys[j, :], mus[i, :])
    return m


def discrete_entropy(ys):
    """
    Compute discrete mutual information.
    """
    num_factors = ys.shape[0]
    h = np.zeros(num_factors)
    for j in range(num_factors):
        h[j] = sklearn.metrics.mutual_info_score(ys[j, :], ys[j, :])
    return h


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import os
from typing import Dict
from typing import Union

# ========================================================================= #
# file hashing                                                              #
# ========================================================================= #


def _yield_file_bytes(file: str, chunk_size=16384):
    with open(file, "rb") as f:
        bytes = True
        while bytes:
            bytes = f.read(chunk_size)
            yield bytes


def _yield_fast_hash_bytes(file: str, chunk_size=16384, num_chunks=3):
    assert num_chunks >= 2
    # return the size in bytes
    size = os.path.getsize(file)
    yield size.to_bytes(length=64 // 8, byteorder="big", signed=False)
    # return file bytes chunks
    if size < chunk_size * num_chunks:
        # we cant return chunks because the file is too small, return everything!
        yield from _yield_file_bytes(file, chunk_size=chunk_size)
    else:
        # includes evenly spaced start, middle and end chunks
        with open(file, "rb") as f:
            for i in range(num_chunks):
                pos = (i * (size - chunk_size)) // (num_chunks - 1)
                f.seek(pos)
                yield f.read(chunk_size)


# ========================================================================= #
# file hashing                                                              #
# ========================================================================= #


def hash_file(file: str, hash_type="md5", hash_mode="full", missing_ok: bool = False) -> str:
    """
    :param file: the path to the file
    :param hash_type: the kind of hash to compute, default is "md5"
    :param hash_mode: "full" uses all the bytes in the file to compute the hash, "fast" uses the start, middle, end bytes as well as the size of the file in the hash.
    :param missing_ok: If enabled, then an error is not thrown if the file is missing, rather an empty hash is returned!
    :return: the hexdigest of the hash
    :raises FileNotFoundError
    """
    import hashlib

    # check the file exists
    if not os.path.isfile(file):
        if missing_ok:
            return ""
        raise FileNotFoundError(f"could not compute hash for missing file: {repr(file)}")
    # get file bytes iterator
    if hash_mode == "full":
        byte_iter = _yield_file_bytes(file=file)
    elif hash_mode == "fast":
        byte_iter = _yield_fast_hash_bytes(file=file)
    else:
        raise KeyError(f"invalid hash_mode: {repr(hash_mode)}")
    # generate hash
    hash = hashlib.new(hash_type)
    for bytes in byte_iter:
        hash.update(bytes)
    hash = hash.hexdigest()
    # done
    return hash


# ========================================================================= #
# file hashing utils                                                        #
# ========================================================================= #


class HashError(Exception):
    """
    Raised if the hash of a file was invalid.
    """


def normalise_hash(hash: Union[str, Dict[str, str]], hash_mode: str) -> str:
    """
    file hashes depend on the mode.
    - Allow hashes to be dictionaries that map the hash_mode to the hash.
      This function returns the correct hash if it is a dictionary.
    """
    if isinstance(hash, dict):
        if hash_mode not in hash:
            raise KeyError(
                f"hash dictionary does not contain a key for the specified mode: {repr(hash_mode)}, available hashes are: {hash}"
            )
        return hash[hash_mode]
    return hash


def validate_file_hash(
    file: str,
    hash: Union[str, Dict[str, str]],
    hash_type: str = "md5",
    hash_mode: str = "full",
    missing_ok: bool = False,
):
    """
    :raises FileNotFoundError, HashError
    """
    hash = normalise_hash(hash=hash, hash_mode=hash_mode)
    # compute the hash
    fhash = hash_file(file=file, hash_type=hash_type, hash_mode=hash_mode, missing_ok=missing_ok)
    # check the hash
    if fhash != hash:
        raise HashError(
            f"computed {hash_mode} {hash_type} hash: {repr(fhash)} does not match expected hash: {repr(hash)} for file: {repr(file)}"
        )


def is_valid_file_hash(
    file: str,
    hash: Union[str, Dict[str, str]],
    hash_type: str = "md5",
    hash_mode: str = "full",
    missing_ok: bool = False,
):
    try:
        validate_file_hash(file=file, hash=hash, hash_type=hash_type, hash_mode=hash_mode, missing_ok=missing_ok)
    except HashError:
        return False
    return True


# ========================================================================= #
# file hashing                                                              #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from functools import lru_cache
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np

from disent.util.iters import LengthIter
from disent.util.visualize.vis_util import get_idx_traversal

# ========================================================================= #
# Types                                                                     #
# ========================================================================= #


NonNormalisedFactorIdx = Union[Sequence[Union[int, str]], Union[int, str]]
NonNormalisedFactorIdxs = Union[Sequence[NonNormalisedFactorIdx], NonNormalisedFactorIdx]
NonNormalisedFactors = Union[np.ndarray, Sequence[Union[int, Sequence]]]


# ========================================================================= #
# Basic State Space                                                         #
# ========================================================================= #


class StateSpace(LengthIter):
    """
    State space where an index corresponds to coordinates (factors/positions) in the factor space.
    ie. State space with multiple factors of variation, where each factor can be a different size.
    """

    def __init__(self, factor_sizes: Sequence[int], factor_names: Optional[Sequence[str]] = None):
        super().__init__()
        # dimension: [read only]
        self.__factor_sizes = np.array(factor_sizes)
        self.__factor_sizes.flags.writeable = False
        # checks
        if self.__factor_sizes.ndim != 1:
            raise ValueError(
                f"`factor_sizes` must be an array with only one dimension, got shape: {self.__factor_sizes.shape}"
            )
        if len(self.__factor_sizes) <= 0:
            raise ValueError(f"`factor_sizes` must be non-empty, got shape: {self.__factor_sizes.shape}")
        # multipliers: [read only]
        self.__factor_multipliers = _dims_multipliers(self.__factor_sizes)
        self.__factor_multipliers.flags.writeable = False
        # total permutations
        self.__size = int(np.prod(factor_sizes))
        # factor names
        self.__factor_names = (
            tuple(f"f{i}" for i in range(self.num_factors)) if (factor_names is None) else tuple(factor_names)
        )
        if len(self.__factor_names) != len(self.__factor_sizes):
            raise ValueError(
                f"Dimensionality mismatch of factor_names and factor_sizes: len({self.__factor_names}) != len({tuple(self.__factor_sizes)})"
            )

    def __len__(self):
        """Same as self.size"""
        return self.size

    def __getitem__(self, idx):
        """Data returned based on the idx"""
        return self.idx_to_pos(idx)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Properties                                                            #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @property
    def size(self) -> int:
        """The number of permutations of factors handled by this state space"""
        return self.__size

    @property
    def num_factors(self) -> int:
        """The number of factors handled by this state space"""
        return len(self.__factor_sizes)

    @property
    def factor_sizes(self) -> np.ndarray:
        """A list of sizes or dimensionality of factors handled by this state space"""
        return self.__factor_sizes

    @property
    def factor_names(self) -> Tuple[str, ...]:
        """A list of names of factors handled by this state space"""
        return self.__factor_names

    @property
    def factor_multipliers(self) -> np.ndarray:
        """
        The cumulative product of the factor_sizes used to convert indices to positions, and positions to indices.
        - The highest values is at the front, the lowest is at the end always being 1.
        - The size of this vector is: num_factors + 1

        Formulas:
            * Use broadcasting to get positions:
                pos = (idx[..., None] % muls[:-1]) // muls[1:]
            * Use broadcasting to get indices
                idx = np.sum(pos * muls[1:], axis=-1)
        """
        return self.__factor_multipliers

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Factor Helpers                                                        #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def normalise_factor_idx(self, factor: NonNormalisedFactorIdx) -> int:
        # convert a factor name to the factor id
        if isinstance(factor, str):
            try:
                f_idx = self.factor_names.index(factor)
            except:
                raise KeyError(f"invalid factor name: {repr(factor)} must be one of: {self.factor_names}")
        else:
            f_idx = int(factor)
        # check that the values are correct
        assert isinstance(f_idx, int)
        assert 0 <= f_idx < self.num_factors
        # return the resulting values
        return f_idx

    def normalise_factor_idxs(self, f_idxs: Optional[NonNormalisedFactorIdxs]) -> np.ndarray:
        # return the default list of factor indices
        if f_idxs is None:
            return np.arange(self.num_factors)
        # normalize a single factor into a list
        if isinstance(f_idxs, (int, str)):
            f_idxs = [f_idxs]
        # convert all the factors to their indices
        f_idxs = np.array([self.normalise_factor_idx(f_idx) for f_idx in f_idxs])
        # done! make sure there are not duplicates!
        assert len(set(f_idxs)) == len(f_idxs), "duplicate factors were found!"
        return f_idxs

    def invert_factor_idxs(self, f_idxs: Optional[NonNormalisedFactorIdxs]) -> np.ndarray:
        f_idxs = self.normalise_factor_idxs(f_idxs)
        # create a mask of factors
        f_mask = np.ones(self.num_factors, dtype="bool")
        f_mask[f_idxs] = False
        # # # select the inverse factors
        return np.where(f_mask)[0]

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Coordinate Transform - any dim array, only last axis counts!          #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def pos_to_idx(self, positions) -> np.ndarray:
        """
        Convert a position to an index (or convert a list of positions to a list of indices)
        - positions are lists of integers, with each element < their corresponding factor size
        - indices are integers < size

        TODO: can factor_multipliers be used to speed this up?
        """
        positions = np.moveaxis(positions, source=-1, destination=0)
        return np.ravel_multi_index(positions, self.__factor_sizes)

    def idx_to_pos(self, indices) -> np.ndarray:
        """
        Convert an index to a position (or convert a list of indices to a list of positions)
        - indices are integers < size
        - positions are lists of integers, with each element < their corresponding factor size

        TODO: can factor_multipliers be used to speed this up?
        """
        positions = np.array(np.unravel_index(indices, self.__factor_sizes))
        return np.moveaxis(positions, source=0, destination=-1)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Iterators                                                             #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def iter_traversal_indices(self, f_idx: int, base_factors):
        base_factors = list(base_factors)
        base_factors[f_idx] = 0
        base_idx = self.pos_to_idx(base_factors)
        step_size = _get_step_size(tuple(self.__factor_sizes), f_idx)
        yield from range(base_idx, base_idx + step_size * self.__factor_sizes[f_idx], step_size)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Sampling Functions - any dim array, only last axis counts!            #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def sample_indices(self, size=None):
        return np.random.randint(0, len(self), size=size)

    def sample_factors(self, size=None, f_idxs: Optional[NonNormalisedFactorIdxs] = None) -> np.ndarray:
        """
        sample randomly from all factors, otherwise the given factor_indices.
        returned values must appear in the same order as factor_indices.

        If factor factor_indices is None, all factors are sampled.
        If size=None then the array returned is the same shape as (len(factor_indices),) or factor_sizes[factor_indices]
        If size is an integer or shape, the samples returned are that shape with the last dimension
            the same size as factor_indices, ie (*size, len(factor_indices))
        """
        # get factor sizes
        if f_idxs is None:
            f_sizes = self.__factor_sizes
        else:
            f_sizes = self.__factor_sizes[self.normalise_factor_idxs(f_idxs)]  # this may be quite slow, add caching?
        # get resample size
        if size is not None:
            # empty np.array(()) gets dtype float which is incompatible with len
            size = np.append(np.array(size, dtype=int), len(f_sizes))
        # sample for factors
        return np.random.randint(0, f_sizes, size=size)

    def sample_missing_factors(
        self, known_factors: NonNormalisedFactors, f_idxs: NonNormalisedFactorIdxs
    ) -> np.ndarray:
        """
        Samples the remaining factors not given in the known_factor_indices.
        ie. fills in the missing values by sampling from the unused dimensions.
        returned values are ordered by increasing factor index and not factor_indices.
        (known_factors must correspond to known_factor_indices)

        - eg. known_factors=[A], known_factor_indices=1
              BECOMES: known_factors=[A], known_factor_indices=[1]
        - eg. known_factors=[A], known_factor_indices=[1]
              = [..., A, ...]
        - eg. known_factors=[[A]], known_factor_indices=[1]
              = [[..., A, ...]]
        - eg. known_factors=[A, B], known_factor_indices=[1, 2]
              = [..., A, B, ...]
        - eg. known_factors=[[A], [B]], known_factor_indices=[1]
              = [[..., A, ...], [..., B, ...]]
        - eg. known_factors=[[A, B], [C, D]], known_factor_indices=[1, 2]
              = [[..., A, B, ...], [..., C, D, ...]]
        """
        f_idxs = self.normalise_factor_idxs(f_idxs)
        f_idxs_inv = self.invert_factor_idxs(f_idxs)
        # normalize shapes
        known_factors = np.array(known_factors)
        # checks
        assert (
            known_factors.ndim >= 1
        ), f"known_factors must have at least one dimension, got shape: {known_factors.shape}"
        assert known_factors.shape[-1] == len(
            f_idxs
        ), f"last dimension of factors must be the same size as the number of f_idxs ({len(f_idxs)}), got shape: {known_factors.shape}"
        # replace the specified factors
        new_factors = np.empty([*known_factors.shape[:-1], self.num_factors], dtype="int")
        new_factors[..., f_idxs] = known_factors
        new_factors[..., f_idxs_inv] = self.sample_factors(size=known_factors.shape[:-1], f_idxs=f_idxs_inv)
        # done!
        return new_factors

    def resample_other_factors(self, factors: NonNormalisedFactors, f_idxs: NonNormalisedFactorIdxs) -> np.ndarray:
        """
        Resample all unspecified factors, keeping f_idxs constant.
        """
        return self.resample_given_factors(factors=factors, f_idxs=self.invert_factor_idxs(f_idxs))

    def resample_given_factors(self, factors: NonNormalisedFactors, f_idxs: NonNormalisedFactorIdxs):
        """
        Resample all specified f_idxs, keeping all remaining factors constant.
        """
        f_idxs = self.normalise_factor_idxs(f_idxs)
        new_factors = np.copy(factors)
        # checks
        assert new_factors.ndim >= 1, f"factors must have at least one dimension, got shape: {new_factors.shape}"
        assert (
            new_factors.shape[-1] == self.num_factors
        ), f"last dimension of factors must be the same size as the number of factors ({self.num_factors}), got shape: {new_factors.shape}"
        # replace the specified factors
        new_factors[..., f_idxs] = self.sample_factors(size=new_factors.shape[:-1], f_idxs=f_idxs)
        # done!
        return new_factors

    def _get_f_idx_and_factors_and_size(
        self,
        f_idx: Optional[int] = None,
        base_factors: Optional[NonNormalisedFactors] = None,
        num: Optional[int] = None,
    ):
        """
        :param f_idx: Sampled randomly in the range [0, num_factors) if not given.
        :param base_factors: Sampled randomly from all possible factors if not given. Coerced into the shape (1, num_factors)
        :param num: Set to the factor size `self.factor_sizes[f_idx]` if not given.
        :return: All values above in a tuple.
        """
        # choose a random factor if not given
        if f_idx is None:
            f_idx = np.random.randint(0, self.num_factors)
        # sample factors if not given
        if base_factors is None:
            base_factors = self.sample_factors(size=1)
        else:
            base_factors = np.reshape(base_factors, (1, self.num_factors))
        # get size if not given
        if num is None:
            num = self.factor_sizes[f_idx]
        else:
            assert num > 0
        # generate a traversal
        base_factors = base_factors.repeat(num, axis=0)
        # return everything
        return f_idx, base_factors, num

    def sample_random_factor_traversal(
        self,
        f_idx: Optional[int] = None,
        base_factors: Optional[NonNormalisedFactors] = None,
        num: Optional[int] = None,
        mode: str = "interval",
        start_index: int = 0,
        return_indices: bool = False,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """
        Sample a single random factor traversal along the
        given factor index, starting from some random base sample.
        """
        f_idx, base_factors, num = self._get_f_idx_and_factors_and_size(f_idx=f_idx, base_factors=base_factors, num=num)
        # generate traversal
        base_factors[:, f_idx] = get_idx_traversal(
            self.factor_sizes[f_idx], num_frames=num, mode=mode, start_index=start_index
        )
        # return factors (num_frames, num_factors)
        if return_indices:
            return base_factors, self.pos_to_idx(base_factors)
        return base_factors

    def sample_random_factor_traversal_grid(
        self,
        num: Optional[int] = None,
        base_factors: Optional[NonNormalisedFactors] = None,
        mode: str = "interval",
        factor_indices: Optional[NonNormalisedFactorIdxs] = None,
        return_indices: bool = False,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        # default values
        if num is None:
            num = int(np.ceil(np.mean(self.factor_sizes)))
        if base_factors is None:
            base_factors = self.sample_factors()
        factor_indices = self.normalise_factor_idxs(factor_indices)
        # generate a grid of factors
        factors_grid = []
        for f_idx in factor_indices:
            factors_grid.append(
                self.sample_random_factor_traversal(
                    f_idx=f_idx, base_factors=base_factors, num=num, mode=mode, start_index=0
                )
            )
        factors_grid = np.stack(factors_grid, axis=0)
        # done!
        if return_indices:
            return factors_grid, self.pos_to_idx(factors_grid)
        return factors_grid


# ========================================================================= #
# Hidden State Space                                                        #
# ========================================================================= #


@lru_cache()
def _get_step_size(factor_sizes, f_idx: int):
    # check values
    assert f_idx >= 0
    assert f_idx < len(factor_sizes)
    # return values
    assert all(f > 0 for f in factor_sizes)
    # return factor size
    pos = np.zeros(len(factor_sizes), dtype="uint8")
    pos[f_idx] = 1
    return int(np.ravel_multi_index(pos, factor_sizes))


def _dims_multipliers(factor_sizes: np.ndarray) -> np.ndarray:
    factor_sizes = np.array(factor_sizes)
    assert factor_sizes.ndim == 1
    return np.append(np.cumprod(factor_sizes[::-1])[::-1], 1)


# @try_njit
# def _idx_to_pos(idxs, dims_mul):
#     factors = np.expand_dims(np.array(idxs, dtype='int'), axis=-1)
#     factors = factors % dims_mul[:-1]
#     factors //= dims_mul[1:]
#     return factors


# ========================================================================= #
# Hidden State Space                                                        #
# ========================================================================= #


# class HiddenStateSpace(_BaseStateSpace):
#
#     """
#     State space where an index corresponds to coordinates (factors/positions) in the factor space.
#     HOWEVER: some factors are treated as hidden/unknown and are thus randomly sampled.
#
#     Inputs to functions act as if known_factor_indices is the new factor space (including factor indexes).
#     Outputs from functions act in the full factor space.
#     """
#
#     def __init__(self, factor_sizes, known_factor_indices=None):
#         if known_factor_indices is None:
#             known_factor_indices = np.arange(len(factor_sizes))
#         factor_sizes, known_factor_indices = np.array(factor_sizes), np.array(known_factor_indices)
#         # known factors indices
#         self._known_factor_indices = known_factor_indices
#         self._known_factor_indices.flags.writeable = False
#         self._known_factor_sizes = factor_sizes[known_factor_indices]
#         # known state space does not include hidden variables, and assumes they are randomly sampled
#         self._known_state_space = StateSpace(self._known_factor_sizes)
#         # full state space includes hidden variables
#         self._full_state_space = StateSpace(factor_sizes)
#
#     @property
#     def size(self):
#         return self._known_state_space.size
#
#     @property
#     def num_factors(self):
#         return self._known_state_space.num_factors
#
#     @property
#     def factor_sizes(self):
#         return self._known_state_space.factor_sizes
#
#     def pos_to_idx(self, positions):
#         positions = np.array(positions)
#         assert 1 <= positions.ndim <= 2, f'positions has incorrect number of dimensions: {positions.ndim}'
#         assert positions.shape[-1] == self._full_state_space.num_factors, 'last dimension of positions must equal the full state space size'
#         # remove the unknown dimensions and return the index
#         return self._known_state_space.pos_to_idx(positions[..., self._known_factor_indices])
#
#     def idx_to_pos(self, indices):
#         indices = np.array(indices)
#         assert 0 <= indices.ndim <= 1, f'indices has incorrect number of dimensions: {indices.ndim}'
#         # get factors and return
#         known_factors = self._known_state_space.idx_to_pos(indices.reshape(-1))
#         sampled_factors = self._full_state_space.sample_missing_factors(known_factors, self._known_factor_indices)
#         return sampled_factors.reshape((*indices.shape, self._full_state_space.num_factors))
#
#     def sample_factors(self, size=None, factor_indices=None):
#         return self._full_state_space.sample_factors(
#             size,
#             self._known_factor_indices[factor_indices] if factor_indices else None
#         )
#
#     def sample_missing_factors(self, known_factors, known_factor_indices):
#         return self._full_state_space.sample_missing_factors(
#             known_factors,
#             self._known_factor_indices[known_factor_indices]
#         )
#
#     def resample_factors(self, factors, fixed_factor_indices):
#         return self._full_state_space.resample_factors(
#             factors,
#             self._known_factor_indices[fixed_factor_indices]
#         )


# ========================================================================= #
# Hidden State Space                                                        #
# ========================================================================= #

# class StateSpaceRemapIndex(object):
#     """Mapping from incorrectly ordered factors to state space indices"""
#
#     def __init__(self, factor_sizes, features):
#         self._states = StateSpace(factor_sizes)
#         # get indices of features
#         orig_indices = self._states.pos_to_idx(features)
#         if np.unique(orig_indices).size != self._states.size:
#             raise ValueError("Features do not cover the entire state space.")
#         # get indices of state space
#         state_indices = np.arange(self._states.size)
#         # mapping
#         self._state_to_orig_idx = np.zeros(self._states.size, dtype=np.int64)
#         self._state_to_orig_idx[orig_indices] = state_indices
#
#     def factors_to_orig_idx(self, factors):
#         """
#         get the original index of factors
#         """
#         return self._state_to_orig_idx[self._states.pos_to_idx(factors)]



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from functools import wraps
from typing import Optional

# ========================================================================= #
# Deprecate                                                                 #
# ========================================================================= #


def _get_traceback_string() -> str:
    import traceback
    from io import StringIO

    # print the stack trace to an in-memory buffer
    file = StringIO()
    traceback.print_stack(file=file)
    return file.getvalue()


def _get_traceback_file_groups():
    # filter the lines
    results = []
    group = []
    for line in _get_traceback_string().splitlines():
        if line.strip().startswith('File "'):
            if group:
                results.append(group)
                group = []
        group.append(line)
    if group:
        results.append(group)
    return results


def _get_stack_file_strings():
    # mimic the output of a traceback so pycharm performs syntax highlighting when printed
    import inspect

    results = []
    for frame_info in inspect.stack():
        results.append(f'File "{frame_info.filename}", line {frame_info.lineno}, in {frame_info.function}')
    return results[::-1]


_TRACEBACK_MODES = {"none", "first", "mini", "traceback"}
DEFAULT_TRACEBACK_MODE = "first"


def deprecated(msg: str, traceback_mode: Optional[str] = None, fn=None):
    """
    Mark a function or class as deprecated, and print a warning the
    first time it is used.
    - This decorator wraps functions, but only replaces the __init__
      method of a class so that we can still inherit from a deprecated class!
    """
    assert isinstance(msg, str), f"msg must be a str, got type: {type(msg)}"
    if traceback_mode is None:
        traceback_mode = DEFAULT_TRACEBACK_MODE
    assert (
        traceback_mode in _TRACEBACK_MODES
    ), f"invalid traceback_mode, got: {repr(traceback_mode)}, must be one of: {sorted(_TRACEBACK_MODES)}"

    def _decorator(fn):
        # we need to handle classes and function separately
        is_class = isinstance(fn, type) and hasattr(fn, "__init__")
        # backup the original function & data
        call_fn = fn.__init__ if is_class else fn
        dat = (fn.__module__, f"{fn.__module__}.{fn.__name__}", str(msg))

        # wrapper function
        @wraps(call_fn)
        def _caller(*args, **kwargs):
            nonlocal dat
            # print the message!
            if dat is not None:
                name, path, dsc = dat
                logger = logging.getLogger(name)
                logger.warning(f"[DEPRECATED] {path} - {repr(dsc)}")
                # get stack trace lines
                if traceback_mode == "first":
                    lines = _get_stack_file_strings()[-3:-2]
                elif traceback_mode == "mini":
                    lines = _get_stack_file_strings()[:-2]
                elif traceback_mode == "traceback":
                    lines = (l[2:] for g in _get_traceback_file_groups()[:-3] for l in g)
                else:
                    lines = []
                # print lines
                for line in lines:
                    logger.warning(f"| {line}")
                # never run this again
                dat = None
            # call the function
            return call_fn(*args, **kwargs)

        # handle function or class
        if is_class:
            fn.__init__ = _caller
        else:
            fn = _caller
        return fn

    # handle function used as decorator, or called directly
    if fn is not None:
        return _decorator(fn)
    else:
        return _decorator


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import warnings
from typing import List
from typing import Optional
from typing import Union

import torch

# ========================================================================= #
# Types                                                                     #
# ========================================================================= #


_DimTypeHint = Optional[Union[int, List[int]]]

_POS_INF = float("inf")
_NEG_INF = float("-inf")

_P_NORM_MAP = {
    "inf": _POS_INF,
    "maximum": _POS_INF,
    "euclidean": 2,
    "manhattan": 1,
    # p < 1 is not a valid norm! but allow it if we set `unbounded_p = True`
    "hamming": 0,
    "minimum": _NEG_INF,
    "-inf": _NEG_INF,
}


# ========================================================================= #
# p-Norm Functions                                                          #
# -- very similar to the generalised mean                                   #
# ========================================================================= #


def torch_dist(xs: torch.Tensor, dim: _DimTypeHint = -1, p: Union[float, str] = 1, keepdim: bool = False):
    """
    Like torch_norm, but allows arbitrary p values that may
    result in the returned values not being a valid norm.
    - norm's require p >= 1
    """
    if isinstance(p, str):
        p = _P_NORM_MAP[p]
    # get absolute values
    xs = torch.abs(xs)
    # compute the specific extreme cases
    # -- its kind of odd that the p-norm and generalised mean converge to the
    #    same values, just from different directions!
    if p == _POS_INF:
        return torch.amax(xs, dim=dim, keepdim=keepdim)
    elif p == _NEG_INF:
        return torch.amin(xs, dim=dim, keepdim=keepdim)
    # get the dimensions
    if dim is None:
        dim = list(range(xs.ndim))
    # warn if the type is wrong
    if p != 1:
        if xs.dtype != torch.float64:
            warnings.warn(
                f"Input tensor to p-norm might not have the required precision, type is {xs.dtype} not {torch.float64}."
            )
    # compute the specific cases
    if p == 0:
        # hamming distance -- number of non-zero entries of the vector
        return torch.sum(xs != 0, dim=dim, keepdim=keepdim, dtype=xs.dtype)
    if p == 1:
        # manhattan distance
        return torch.sum(xs, dim=dim, keepdim=keepdim)
    else:
        # p-norm (if p==2, then euclidean distance)
        return torch.sum(xs**p, dim=dim, keepdim=keepdim) ** (1 / p)


def torch_norm(xs: torch.Tensor, dim: _DimTypeHint = -1, p: Union[float, str] = 1, keepdim: bool = False):
    """
    Compute the generalised p-norm over the given dimension of a vector!
    - p values must be >= 1

    Closely related to the generalised mean.
        - p-norm:          (sum(|x|^p)) ^ (1/p)
        - gen-mean:  (1/n * sum(x ^ p)) ^ (1/p)
    """
    if isinstance(p, str):
        p = _P_NORM_MAP[p]
    # check values
    if p < 1:
        raise ValueError(
            f"p-norm cannot have a p value less than 1, got: {repr(p)}, to bypass this error set `unbounded_p=True`."
        )
    # return norm
    return torch_dist(xs=xs, dim=dim, p=p, keepdim=keepdim)


def torch_norm_euclidean(xs, dim: _DimTypeHint = -1, keepdim: bool = False):
    return torch_dist(xs, dim=dim, p="euclidean", keepdim=keepdim)


def torch_norm_manhattan(xs, dim: _DimTypeHint = -1, keepdim: bool = False):
    return torch_dist(xs, dim=dim, p="manhattan", keepdim=keepdim)


def torch_dist_hamming(xs, dim: _DimTypeHint = -1, keepdim: bool = False):
    return torch_dist(xs, dim=dim, p="hamming", keepdim=keepdim)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from copy import deepcopy
from typing import Optional
from typing import Sequence

import hydra
from omegaconf import DictConfig
from omegaconf import ListConfig
from omegaconf import OmegaConf

from disent.util.deprecate import deprecated

log = logging.getLogger(__name__)


# ========================================================================= #
# Helper                                                                    #
# ========================================================================= #


def make_non_strict(cfg: DictConfig):
    """
    Convert the config into a mutable version.
    """
    cfg = deepcopy(cfg)
    return OmegaConf.create({**cfg})


# ========================================================================= #
# END                                                                       #
# ========================================================================= #
